--========================================================--
--                Chat tab bind                           --
--                                                        --
-- Author      :  kurapica125@outlook.com                 --
-- Create Date :  2017/11/19                              --
--========================================================--

--========================================================--
Scorpio            "ChatTabBind"                      "1.0.0"
--========================================================--

SpecialFrameFix = {}
FrameMap = {}
SecureInited = {}

----------------------------------------------
--------------- Choose Frame Mask ------------
----------------------------------------------
local _MaskMode = false
local _ChooseFrame
local _MouseFocusInitFrame
local _MouseFocusFrame

local _Mask = CreateFrame("Button")
_Mask:Hide()
_Mask:SetToplevel(true)
_Mask:SetFrameStrata("TOOLTIP")
_Mask:EnableMouse(true)
_Mask:EnableMouseWheel(true)
_Mask:RegisterForClicks("AnyUp")
_Mask:SetBackdrop{
    bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
    edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
    tile = true, tileSize = 16, edgeSize = 8,
    insets = { left = 3, right = 3, top = 3, bottom = 3 }
}
_Mask:SetBackdropColor(0, 1, 0, 0.8)

_Mask:SetScript("OnClick", function(self, btn)
    if btn == "LeftButton" then
        _ChooseFrame = _MouseFocusFrame
    else
        _ChooseFrame = nil
    end
    _MaskMode = nil
end)

_Mask:SetScript("OnMouseWheel", function(self, wheel)
    if wheel > 0 then
        if _MouseFocusFrame then
            local parent = _MouseFocusFrame:GetParent()
            if parent and parent ~= UIParent and parent ~= WorldFrame and parent:GetName() then
                _MouseFocusFrame = parent
                ShowGameTooltip()
            end
        end
    else
        if _MouseFocusInitFrame then
            _MouseFocusFrame = _MouseFocusInitFrame
            ShowGameTooltip()
        end
    end
end)

----------------------------------------------
-------------- Addon Event Handler -----------
----------------------------------------------
function OnLoad(self)
    _SVDB = SVManager.SVCharManager("ChatTabBind_DB")
    _SVDB:SetDefault("LabelMap", {})
end

function OnQuit(self)
    wipe(_SVDB.LabelMap)

    for k, m in pairs(FrameMap) do
        _SVDB.LabelMap[k:GetName()] = m.Label
        FCF_Close(m.ChatFrame)
    end
end

----------------------------------------------
-------------- Addon Slash Command -----------
----------------------------------------------
__SlashCmd__"chatbind" __SlashCmd__ "ctb"
__Thread__()
function ChooseFrame(name)
    if not name or strtrim(name) == "" then
        return print(_Locale["/chatbind name - please input the chat tab's label"])
    end

    name = strtrim(name)

    for k, m in pairs(FrameMap) do
        if name == m.Label then
            return print(_Locale["/chatbind name - the name is already used"])
        end
    end

    _MaskMode = true
    _ChooseFrame = nil
    _MouseFocusInitFrame = nil
    _MouseFocusFrame = nil

    while _MaskMode and not InCombatLockdown() do
        local frame = GetMouseFocus()

        if frame ~= _Mask then
            while frame and not frame:GetName() do
                frame = frame:GetParent()
            end

            if _MouseFocusInitFrame ~= frame then
                if frame == UIParent or frame == WorldFrame then
                    if _MouseFocusInitFrame then
                        _MouseFocusInitFrame = nil
                        _MouseFocusFrame = nil
                        _Mask:ClearAllPoints()
                        _Mask:Hide()
                        _Mask:SetParent(nil)
                        HideGameTooltip()
                    end
                else
                    _MouseFocusInitFrame = frame
                    _MouseFocusFrame = frame
                    _Mask:SetParent(frame)
                    _Mask:SetAllPoints(frame)
                    _Mask:Show()
                    ShowGameTooltip()
                end
            end
        end

        Next()
    end

    _MaskMode = false
    _Mask:Hide()
    HideGameTooltip()

    if _ChooseFrame then
        BindFrameToChatFrame(_ChooseFrame, name)
    end
end

----------------------------------------------
-------------- System Event Handler ----------
----------------------------------------------
__Thread__() __SystemEvent__()
function PLAYER_ENTERING_WORLD(self)
    _M:UnregisterEvent("PLAYER_ENTERING_WORLD")

    Delay(2)

    local map = _SVDB.LabelMap
    while next(map) do
        Delay(0.1)
        local tab = _G.SELECTED_CHAT_FRAME

        for k, v in pairs(map) do
            if v and _G[k] and _G[k].GetName and _G[k]:GetName() == k then
                BindFrameToChatFrame(_G[k], v)
                map[k] = nil
            end
        end

        FCF_Tab_OnClick(tab, "LeftButton")

        if not next(map) then return end

        Event("ADDON_LOADED")
    end
end

----------------------------------------------
-------------- Secure Hook Handler -----------
----------------------------------------------
__SecureHook__ "FCF_Close"
function UnbindChatFrame(frame, fallback)
    if ( fallback ) then
        frame=fallback
    end
    for f, m in pairs(FrameMap) do
        if m.ChatFrame == frame then
            FrameMap[f] = nil

            f:SetParent(m.OriginalPar)

            f:ClearAllPoints()
            for i, anchor in ipairs(m.OriginalLoc) do
                f:SetPoint(unpack(anchor))
            end

            f:SetSize(unpack(m.OriginalSize))

            RunFixCode(f, false)

            break
        end
    end
end

----------------------------------------------
------------------ Addon Helper --------------
----------------------------------------------
function ShowGameTooltip()
    if _Mask:IsVisible() and _MouseFocusFrame then
        GameTooltip:SetOwner(_Mask, "ANCHOR_BOTTOMRIGHT")
        GameTooltip:SetText(_MouseFocusFrame:GetName())
        GameTooltip:Show()
    end
end

function HideGameTooltip()
    GameTooltip:Hide()
end

function BindFrameToChatFrame(frame, name)
    local tabStatus = {}
    local canAdd = false

    for i=1, NUM_CHAT_WINDOWS do
        local _, _, _, _, _, _, shown = FCF_GetChatWindowInfo(i)
        local chatFrame = _G["ChatFrame"..i]

        tabStatus[i] = (not shown and not chatFrame.isDocked)
        if tabStatus[i] then canAdd = true end
    end

    if not canAdd then
        return print(_Locale["Can't create new chat tab."])
    end

    FCF_OpenNewWindow(name)

    print(_Locale["Add Chat tab %q for %q"]:format(name, frame:GetName()))

    for i=1, NUM_CHAT_WINDOWS do
        local _, _, _, _, _, _, shown = FCF_GetChatWindowInfo(i)
        local chatFrame = _G["ChatFrame"..i]

        if tabStatus[i] and shown and chatFrame.name == name then
            -- clear stale messages
            chatFrame:Clear()

            -- Listen to the standard messages
            ChatFrame_RemoveAllMessageGroups(chatFrame)
            ChatFrame_RemoveAllChannels(chatFrame)
            ChatFrame_ReceiveAllPrivateMessages(chatFrame)

            -- Save Location
            local loc = {}

            for i = 1, frame:GetNumPoints() do
                loc[i] = { frame:GetPoint(i) }
            end

            FrameMap[frame] = {
                Label       = chatFrame.name,
                ChatFrame   = chatFrame,
                OriginalPar = frame:GetParent(),
                OriginalLoc = loc,
                OriginalSize= { frame:GetSize() },
            }

            frame:SetParent(chatFrame)
            frame:ClearAllPoints()
            frame:SetAllPoints()

            RunFixCode(frame, true)

            if not SecureInited[chatFrame] then
                SecureInited[chatFrame] = true
                chatFrame:HookScript("OnShow", ChatFrame_OnShow)
            end
        end
    end
end

function ChatFrame_OnShow(self)
    for f, m in pairs(FrameMap) do
        if m.ChatFrame == self then
            f:SetParent(self)
            f:ClearAllPoints()
            f:SetAllPoints()

            RunFixCode(f, true)

            break
        end
    end
end

function RunFixCode(frame, asbind)
    local name = frame:GetName()

    for pattern, fix in pairs(SpecialFrameFix) do
        if name:match(pattern) then
            return fix(frame, asbind)
        end
    end
end
